@@ -1,17 +1,17 @@
 =begin
-Copyright (c) 2012, Brice Videau <brice.videau@imag.fr>
-Copyright (c) 2012, Vincent Danjean <Vincent.Danjean@ens-lyon.org>
+Copyright (c) 2012-2020, Brice Videau <bvideau@anl.gov>
+Copyright (c) 2012-2020, Vincent Danjean <Vincent.Danjean@ens-lyon.org>
 All rights reserved.
-      
+
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
-    
+
 1. Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
-        
+
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -27,6 +27,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 require 'yaml'
 
 module IcdGenerator
+
   $api_entries = {}
   $key_entries = {}
   $api_entries_array = []
@@ -55,7 +56,7 @@ module IcdGenerator
   $specific_loader_funcs = ["clGetExtensionFunctionAddress","clGetPlatformIDs",
                          "clGetGLContextInfoKHR", "clUnloadCompiler",
     "clCreateContext", "clCreateContextFromType", "clWaitForEvents"]
-  $header_files = ["/usr/include/CL/cl.h", "/usr/include/CL/cl_gl.h",
+  $header_files = ["/usr/include/CL/cl.h", "/usr/include/CL/cl_gl.h", "/usr/include/CL/cl_egl.h",
     "/usr/include/CL/cl_ext.h", "/usr/include/CL/cl_gl_ext.h"]
   $windows_header_files = ["/usr/include/CL/cl_dx9_media_sharing.h", "/usr/include/CL/cl_d3d11.h", "/usr/include/CL/cl_d3d10.h"]
   $cl_data_type_error = { "cl_platform_id"   => "CL_INVALID_PLATFORM",
@@ -71,19 +72,19 @@ module IcdGenerator
   $versions_entries = []
   $buff=50
   $license = <<EOF
-Copyright (c) 2012, Brice Videau <brice.videau@imag.fr>
-Copyright (c) 2012, Vincent Danjean <Vincent.Danjean@ens-lyon.org>
+Copyright (c) 2012-2020, Brice Videau <bvideau@anl.gov>
+Copyright (c) 2012-2020, Vincent Danjean <Vincent.Danjean@ens-lyon.org>
 All rights reserved.
-      
+
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
-    
+
 1. Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
-        
+
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -111,7 +112,7 @@ EOF
     }
     api_entries.each{ |entry|
 #      puts entry
-      begin 
+      begin
         entry_name = entry.match(/CL_API_CALL(.*?)\(/m)[1].strip
       rescue
         entry_name = entry.match(/(\S*?)\(/m)[1].strip
@@ -170,12 +171,30 @@ EOF
     return headers
   end
 
+  def self.include_headers_no_warnings
+    headers_no_warn = ""
+    headers_no_warn << "#pragma GCC diagnostic push\n"
+    headers_no_warn << "#  pragma GCC diagnostic ignored \"-Wcpp\"\n"
+    headers_no_warn << "#  define CL_USE_DEPRECATED_OPENCL_1_0_APIS\n"
+    headers_no_warn << "#  define CL_USE_DEPRECATED_OPENCL_1_1_APIS\n"
+    headers_no_warn << "#  define CL_USE_DEPRECATED_OPENCL_1_2_APIS\n"
+    headers_no_warn << "#  define CL_USE_DEPRECATED_OPENCL_2_0_APIS\n"
+    headers_no_warn << "#  define CL_USE_DEPRECATED_OPENCL_2_1_APIS\n"
+    headers_no_warn << "#  define CL_USE_DEPRECATED_OPENCL_2_2_APIS\n"
+    headers_no_warn << "#  define CL_TARGET_OPENCL_VERSION 300\n"
+    headers_no_warn << "#  include <CL/opencl.h>\n"
+    headers_no_warn << self.include_headers
+    headers_no_warn << "#pragma GCC diagnostic pop\n"
+    headers_no_warn
+   end
+
   ##########################################################
   ##########################################################
   # generate mode
   def self.generate_libdummy_icd_header
     libdummy_icd_structures = "/**\n#{$license}\n*/\n"
-    libdummy_icd_structures +=  "#include <CL/opencl.h>\n"
+    libdummy_icd_structures += "#define CL_TARGET_OPENCL_VERSION 300\n"
+    libdummy_icd_structures += "#include <CL/opencl.h>\n"
     libdummy_icd_structures += self.include_headers
     libdummy_icd_structures += "\n\nstruct _cl_icd_dispatch;\n"
     libdummy_icd_structures += "struct _cl_platform_id { struct _cl_icd_dispatch *dispatch; };\n\n"
@@ -189,7 +208,7 @@ EOF
     }
     libdummy_icd_structures += "};\n\n"
     libdummy_icd_structures += "#pragma GCC visibility push(hidden)\n\n"
-    libdummy_icd_structures += "struct _cl_icd_dispatch master_dispatch; \n\n"
+    libdummy_icd_structures += "extern struct _cl_icd_dispatch master_dispatch; \n\n"
     $use_name_in_test.each { |k, f|
       libdummy_icd_structures += "typeof(#{f}) INT#{f};\n"
     }
@@ -209,7 +228,7 @@ EOF
     comma=","
     ($api_entries.length+$buff).times { |i|
       comma="" if (i == $api_entries.length+$buff-1)
-      if( $use_name_in_test[i] ) then 
+      if( $use_name_in_test[i] ) then
         libdummy_icd_source += "  (void(*)(void))& INT#{$known_entries[i]}#{comma}\n"
       else
         libdummy_icd_source += "  (void(*)(void))& dummyFunc#{i}#{comma}\n"
@@ -218,19 +237,12 @@ EOF
     libdummy_icd_source += "};\n"
     return libdummy_icd_source
   end
-  
+
   def self.generate_run_dummy_icd_source
     run_dummy_icd = "/**\n#{$license}\n*/\n"
     run_dummy_icd += "#include <stdlib.h>\n"
     run_dummy_icd += "#include <stdio.h>\n"
-    run_dummy_icd += "#pragma GCC diagnostic push\n"
-    run_dummy_icd += "#  pragma GCC diagnostic ignored \"-Wcpp\"\n"
-    run_dummy_icd += "#  define CL_USE_DEPRECATED_OPENCL_1_0_APIS\n"
-    run_dummy_icd += "#  define CL_USE_DEPRECATED_OPENCL_1_1_APIS\n"
-    run_dummy_icd += "#  define CL_USE_DEPRECATED_OPENCL_1_2_APIS\n"
-    run_dummy_icd += "#  include <CL/opencl.h>\n"
-    run_dummy_icd += self.include_headers
-    run_dummy_icd += "#pragma GCC diagnostic pop\n"
+    run_dummy_icd += self.include_headers_no_warnings
     run_dummy_icd += "\n\n"
     $api_entries.each_key { |func_name|
        next if $forbidden_funcs.include?(func_name)
@@ -251,6 +263,8 @@ EOF
          run_dummy_icd += "  #{func_name}(1,(cl_event*)&chosen_platform);\n"
        elsif func_name == "clGetExtensionFunctionAddress" then
          run_dummy_icd += "  #{func_name}(\"extLIG\");\n"
+       elsif func_name == "clGetExtensionFunctionAddressForPlatform" then
+         run_dummy_icd += "  #{func_name}((cl_platform_id )chosen_platform,\"extLIG\");\n"
        elsif func_name == "clUnloadCompiler" then
          run_dummy_icd += "  #{func_name}();\n"
        else
@@ -289,14 +303,7 @@ EOF
 #include <dlfcn.h>
 
 EOF
-    run_dummy_icd_weak += "#pragma GCC diagnostic push\n"
-    run_dummy_icd_weak += "#  pragma GCC diagnostic ignored \"-Wcpp\"\n"
-    run_dummy_icd_weak += "#  define CL_USE_DEPRECATED_OPENCL_1_0_APIS\n"
-    run_dummy_icd_weak += "#  define CL_USE_DEPRECATED_OPENCL_1_1_APIS\n"
-    run_dummy_icd_weak += "#  define CL_USE_DEPRECATED_OPENCL_1_2_APIS\n"
-    run_dummy_icd_weak += "#  include <CL/opencl.h>\n"
-    run_dummy_icd_weak += self.include_headers
-    run_dummy_icd_weak += "#pragma GCC diagnostic pop\n"
+    run_dummy_icd_weak += self.include_headers_no_warnings
 
     $api_entries.each { |func_name, entry|
       next if $noweak_funcs.include?(func_name)
@@ -319,7 +326,7 @@ EOF
         p = p.split
         p = p[-1].gsub("*","")
       }
-       
+
       run_dummy_icd_weak += "  if(p) {\n"
       run_dummy_icd_weak += "    return((*(typeof(#{func_name})*)p)("
       run_dummy_icd_weak += ps.join(", ")
@@ -333,8 +340,96 @@ EOF
     return run_dummy_icd_weak
   end
 
+  def self.generate_icd_layer_source
+    icd_layer_source = "/**\n#{$license}\n*/\n"
+    icd_layer_source += <<EOF
+#include <stdio.h>
+#include "ocl_icd_layer.h"
+
+static struct _cl_icd_dispatch dispatch = {NULL};
+static const struct _cl_icd_dispatch *tdispatch;
+
+CL_API_ENTRY cl_int CL_API_CALL
+clGetLayerInfo(
+    cl_layer_info  param_name,
+    size_t         param_value_size,
+    void          *param_value,
+    size_t        *param_value_size_ret) {
+  if (param_value_size && !param_value)
+    return CL_INVALID_VALUE;
+  if (!param_value && !param_value_size_ret)
+    return CL_INVALID_VALUE;
+  switch (param_name) {
+  case CL_LAYER_API_VERSION:
+    if (param_value_size < sizeof(cl_layer_api_version))
+      return CL_INVALID_VALUE;
+    if (param_value)
+      *((cl_layer_api_version *)param_value) = CL_LAYER_API_VERSION_100;
+    if (param_value_size_ret)
+      *param_value_size_ret = sizeof(cl_layer_api_version);
+    break;
+  default:
+    return CL_INVALID_VALUE;
+  }
+  return CL_SUCCESS;
+}
+
+static void _init_dispatch(void);
+
+CL_API_ENTRY cl_int CL_API_CALL
+clInitLayer(
+    cl_uint                         num_entries,
+    const struct _cl_icd_dispatch  *target_dispatch,
+    cl_uint                        *num_entries_out,
+    const struct _cl_icd_dispatch **layer_dispatch) {
+  if (!target_dispatch || !layer_dispatch ||!num_entries_out || num_entries < OCL_ICD_LAST_FUNCTION+1)
+    return -1;
+
+  _init_dispatch();
+
+  tdispatch = target_dispatch;
+  *layer_dispatch = &dispatch;
+  *num_entries_out = OCL_ICD_LAST_FUNCTION+1;
+  return CL_SUCCESS;
+}
+
+EOF
+
+    $api_entries.each { |func_name, entry|
+      clean_entry = entry.sub(/(.*\)).*/m,'\1').gsub("/*","").gsub("*/","").gsub("\r","") + "{\n"
+      return_type = entry.match(/CL_API_ENTRY (.*) CL_API_CALL/)[1]
+      parameters = clean_entry.match(/\(.*\)/m)[0][1..-2]
+      parameters.gsub!(/\[.*?\]/,"")
+      parameters.sub!(/\(.*?\*\s*(.*?)\)\s*\(.*?\)/m,'\1')
+      clean_entry = clean_entry.gsub(func_name, func_name+"_wrap").gsub("CL_API_ENTRY", "").gsub("CL_API_CALL", "")
+      icd_layer_source += "static "
+      icd_layer_source += clean_entry.gsub(/\*\[.*?\]/,"*  ").gsub(/\[.+?\]/,"")
+
+      ps = parameters.split(",")
+      ps = ps.collect { |p|
+        p = p.split
+        p = p[-1].gsub("*","")
+      }
+      icd_layer_source += "  printf(\"#{func_name}\\n\");\n"
+      icd_layer_source += "  return tdispatch->#{func_name}("
+      if func_name != "clUnloadCompiler"
+        icd_layer_source += ps.join(", ")
+      end
+      icd_layer_source += ");\n"
+      icd_layer_source += "}\n\n"
+    }
+
+    icd_layer_source += "static void _init_dispatch(void) {\n"
+    $api_entries.each { |func_name, _|
+      icd_layer_source += "  dispatch.#{func_name} = &#{func_name}_wrap;\n"
+    }
+    icd_layer_source += "}"
+
+    icd_layer_source
+  end
+
   def self.generate_sources(from_headers=true, from_database=false, database=nil)
-    if from_headers then        
+    if from_headers then
       parse_headers
     end
     if from_database then
@@ -352,6 +447,9 @@ EOF
     File.open('run_dummy_icd_weak_gen.c','w') { |f|
       f.puts generate_run_dummy_icd_weak_source
     }
+    File.open('dummy_icd_layer_gen.c','w') { |f|
+      f.puts generate_icd_layer_source
+    }
   end
 
   ##########################################################
@@ -361,28 +459,24 @@ EOF
     ocl_icd_header = "/**\n#{$license}\n*/\n\n"
     ocl_icd_header += "#ifndef OCL_ICD_H\n"
     ocl_icd_header += "#define OCL_ICD_H\n"
-    ocl_icd_header += "#pragma GCC diagnostic push\n"
-    ocl_icd_header += "#  pragma GCC diagnostic ignored \"-Wcpp\"\n"
-    ocl_icd_header += "#  define CL_USE_DEPRECATED_OPENCL_1_0_APIS\n"
-    ocl_icd_header += "#  define CL_USE_DEPRECATED_OPENCL_1_1_APIS\n"
-    ocl_icd_header += "#  define CL_USE_DEPRECATED_OPENCL_1_2_APIS\n"
-    ocl_icd_header += "#  include <CL/opencl.h>\n"
-    ocl_icd_header += self.include_headers
-    ocl_icd_header += "#pragma GCC diagnostic pop\n"
+    ocl_icd_header += self.include_headers_no_warnings
     ocl_icd_header += <<EOF
 
 #define OCL_ICD_API_VERSION	1
 #define OCL_ICD_IDENTIFIED_FUNCTIONS	#{$known_entries.count}
+#define OCL_ICD_LAST_FUNCTION	#{$known_entries.keys.max}
 
 struct _cl_icd_dispatch {
 EOF
     nb=0
     $api_entries_array.each { |entry|
       version = entry.split("\n").
-	grep(/ CL_API_SUFFIX__(VERSION_[0-9_]+)[^0-9_]/).join('').
-	gsub(/.* CL_API_SUFFIX__(VERSION_[0-9_]+)[^0-9_].*$/, '\1')
+	grep(/ CL_(API|EXT)_SUFFIX__(VERSION_[0-9_]+)[^0-9_]/).join('').
+	gsub(/.* CL_(API|EXT)_SUFFIX__(VERSION_[0-9_]*[0-9])_*[^0-9_].*$/, '\2')
       if (version != '') then
 	ocl_icd_header += '#ifdef CL_'+version+"\n"
+      elsif not entry =~ / clUnknown/
+        ocl_icd_header += "#warning no version\n"
       end
       ocl_icd_header += entry.gsub("\r","").
 	sub(/CL_API_CALL\n?(.*?)\(/m,'(CL_API_CALL*\1)('+"\n  ").
@@ -454,13 +548,13 @@ EOF
     }
     return ocl_icd_loader_map
   end
- 
+
   def self.generate_ocl_icd_bindings_source
     ocl_icd_bindings_source = "/**\n#{$license}\n*/\n"
     ocl_icd_bindings_source += "#include \"ocl_icd.h\"\n"
     ocl_icd_bindings_source += "struct _cl_icd_dispatch master_dispatch = {\n"
     ($api_entries.length+$buff-1).times { |i|
-      if( $known_entries[i] ) then 
+      if( $known_entries[i] ) then
         ocl_icd_bindings_source += "  #{$known_entries[i]},\n"
       else
         ocl_icd_bindings_source += "  (void *) NULL,\n"
@@ -474,8 +568,8 @@ EOF
     ocl_icd_bindings_source += "};\n"
     ocl_icd_bindings_source += <<EOF
 
-CL_API_ENTRY cl_int CL_API_CALL clIcdGetPlatformIDsKHR(  
-             cl_uint num_entries, 
+CL_API_ENTRY cl_int CL_API_CALL clIcdGetPlatformIDsKHR(
+             cl_uint num_entries,
              cl_platform_id *platforms,
              cl_uint *num_platforms) {
   if( platforms == NULL && num_platforms == NULL )
@@ -503,9 +597,9 @@ CL_API_ENTRY void * CL_API_CALL clGetExtensionFunctionAddress(
   return NULL;
 }
 CL_API_ENTRY cl_int CL_API_CALL clGetPlatformInfo(
-             cl_platform_id   platform, 
+             cl_platform_id   platform,
              cl_platform_info param_name,
-             size_t           param_value_size, 
+             size_t           param_value_size,
              void *           param_value,
              size_t *         param_value_size_ret) CL_API_SUFFIX__VERSION_1_0 {
 #error You ahve to fill this function with your information or assert that your version responds to CL_PLATFORM_ICD_SUFFIX_KHR
@@ -569,19 +663,42 @@ EOF
     return ocl_icd_bindings_source
   end
 
+  def self.generate_get_extension_address_for_platform
+    src = <<EOF
+  if( func_name == NULL )
+    return NULL;
+  struct func_desc const * fn=&function_description[0];
+  int lenfn=strlen(func_name);
+  if (lenfn > 3 &&
+      (strcmp(func_name+lenfn-3, "KHR")==0 || strcmp(func_name+lenfn-3, "EXT")==0)) {
+    while (fn->name != NULL) {
+      if (strcmp(func_name, fn->name)==0)
+        RETURN(fn->addr);
+      fn++;
+    }
+  }
+EOF
+    return src
+  end
+
   def self.generate_ocl_icd_loader_gen_source
     skip_funcs = $specific_loader_funcs
     ocl_icd_loader_gen_source = "/**\n#{$license}\n*/\n"
+    ocl_icd_loader_gen_source += "#include <string.h>\n"
     ocl_icd_loader_gen_source += "#include \"ocl_icd_loader.h\"\n"
     ocl_icd_loader_gen_source += "#define DEBUG_OCL_ICD_PROVIDE_DUMP_FIELD\n"
     ocl_icd_loader_gen_source += "#include \"ocl_icd_debug.h\"\n"
-    $api_entries.each { |func_name, entry|
+    api_proc = proc { |disp, (func_name, entry)|
       next if skip_funcs.include?(func_name)
       clean_entry = entry.sub(/(.*\)).*/m,'\1').gsub("/*","").gsub("*/","").gsub("\r","") + "{\n"
       return_type = entry.match(/CL_API_ENTRY (.*) CL_API_CALL/)[1]
       parameters = clean_entry.match(/\(.*\)/m)[0][1..-2]
       parameters.gsub!(/\[.*?\]/,"")
       parameters.sub!(/\(.*?\*\s*(.*?)\)\s*\(.*?\)/m,'\1')
+      if disp
+        clean_entry = clean_entry.gsub(func_name, func_name+"_disp").gsub("CL_API_ENTRY", "").gsub("CL_API_CALL", "")
+        ocl_icd_loader_gen_source +=  '__attribute__((visibility("hidden"))) '
+      end
       ocl_icd_loader_gen_source += clean_entry.gsub(/\*\[.*?\]/,"*  ").gsub(/\[.+?\]/,"")
       first_parameter = parameters.match(/.*?\,/m)
       if not first_parameter then
@@ -590,13 +707,21 @@ EOF
         first_parameter = first_parameter[0][0..-2]
       end
       fps = first_parameter.split
-      ocl_icd_loader_gen_source += "  debug_trace();\n"
-      raise "Unsupported data_type #{fps[0]}" if not $cl_data_type_error[fps[0]]
       ps = parameters.split(",")
       ps = ps.collect { |p|
         p = p.split
         p = p[-1].gsub("*","")
       }
+      if !disp
+        ocl_icd_loader_gen_source += "  debug_trace();\n"
+        ocl_icd_loader_gen_source += "  _initClIcd_no_inline();\n" if fps[0] == "cl_platform_id"
+        ocl_icd_loader_gen_source += "  if (__builtin_expect (!!_first_layer, 0))\n"
+        ocl_icd_loader_gen_source += "    return _first_layer->dispatch.#{func_name}("
+        ocl_icd_loader_gen_source += ps.join(", ")
+        ocl_icd_loader_gen_source += ");\n"
+      end
+      ocl_icd_loader_gen_source += generate_get_extension_address_for_platform if func_name == "clGetExtensionFunctionAddressForPlatform"
+      raise "Unsupported data_type #{fps[0]}" if ! $cl_data_type_error[fps[0]]
       error_handler = lambda {
          if(ps.include?("errcode_ret")) then
           ocl_icd_loader_gen_source += "    if( errcode_ret != NULL ) {\n";
@@ -621,7 +746,7 @@ EOF
           end
         end
       }
-       
+
       if(fps[0] == "cl_platform_id") then
         ocl_icd_loader_gen_source += "  #{fps[1]}=selectPlatformID(#{fps[1]});\n"
       end
@@ -638,6 +763,8 @@ EOF
       ocl_icd_loader_gen_source += "));\n"
       ocl_icd_loader_gen_source += "}\n\n"
     }
+    $api_entries.each &(api_proc.curry[true])
+    $api_entries.each &(api_proc.curry[false])
     ocl_icd_loader_gen_source += "#pragma GCC visibility push(hidden)\n\n"
     skip_funcs = $specific_loader_funcs
     $api_entries.each { |func_name, entry|
@@ -645,6 +772,7 @@ EOF
       #next if func_name.match(/KHR$/)
       if (skip_funcs.include?(func_name)) then
         ocl_icd_loader_gen_source += "extern typeof(#{func_name}) #{func_name}_hid;\n"
+        ocl_icd_loader_gen_source += "extern typeof(#{func_name}) #{func_name}_disp;\n"
       else
         ocl_icd_loader_gen_source += "typeof(#{func_name}) #{func_name}_hid __attribute__ ((alias (\"#{func_name}\"), visibility(\"hidden\")));\n"
       end
@@ -653,12 +781,30 @@ EOF
     $api_entries.each { |func_name, entry|
       #next if func_name.match(/EXT$/)
       #next if func_name.match(/KHR$/)
-      ocl_icd_loader_gen_source += "  {\"#{func_name}\", (void(* const)(void))&#{func_name}_hid },\n"
+      ocl_icd_loader_gen_source << "  {\"#{func_name}\", (void(* const)(void))&#{func_name}_hid },\n"
     }
-    ocl_icd_loader_gen_source += <<EOF
+    ocl_icd_loader_gen_source << <<EOF
   {NULL, NULL}
 };
 
+EOF
+    ocl_icd_loader_gen_source << "struct _cl_icd_dispatch master_dispatch = {\n"
+    ($api_entries.length+$buff-1).times { |i|
+      if( $known_entries[i] ) then
+        ocl_icd_loader_gen_source << "  #{$known_entries[i]}_disp,\n"
+      else
+        ocl_icd_loader_gen_source << "  (void *) NULL,\n"
+      end
+    }
+    if( $known_entries[$api_entries.length+$buff-1] ) then
+      ocl_icd_loader_gen_source << "  #{$known_entries[$api_entries.length+$buff-1]}\n"
+    else
+      ocl_icd_loader_gen_source << "  (void *) NULL\n"
+    end
+    ocl_icd_loader_gen_source << "};\n"
+
+    ocl_icd_loader_gen_source << <<EOF
+
 #ifdef DEBUG_OCL_ICD
 void dump_platform(clGEFA_t f, cl_platform_id pid) {
   debug(D_ALWAYS, "platform @%p:  name=field_in_struct [clGetExtensionFunctionAddress(name)/clGetExtensionFunctionAddressForPlatform(name)]", pid);
@@ -666,10 +812,10 @@ EOF
     $api_entries_array.each { |entry|
       e = entry.gsub("\r"," ").gsub("\n"," ").gsub("\t"," ").
         sub(/.*CL_API_CALL *([^ ()]*)[ ()].*$/m, '\1')
-      ocl_icd_loader_gen_source += "  dump_field(pid, f, #{e});\n"
+      ocl_icd_loader_gen_source << "  dump_field(pid, f, #{e});\n"
     }
 
-    ocl_icd_loader_gen_source += <<EOF
+    ocl_icd_loader_gen_source << <<EOF
 }
 #endif
 
@@ -678,7 +824,7 @@ EOF
 EOF
     return ocl_icd_loader_gen_source;
   end
-  
+
   def self.generate_from_database(yamlfile)
     load_database(yamlfile)
     File.open('ocl_icd.h','w') { |f|
@@ -696,6 +842,9 @@ EOF
     File.open('ocl_icd_loader_gen.c','w') { |f|
       f.puts generate_ocl_icd_loader_gen_source
     }
+    File.open('dummy_icd_layer_gen.c','w') { |f|
+      f.puts generate_icd_layer_source
+    }
   end
 
   ##########################################################
@@ -775,7 +924,7 @@ OptionParser.new do |opts|
   do |v|
     options[:input] = v
   end
-  opts.on("-s", "--[no-]system-headers", 
+  opts.on("-s", "--[no-]system-headers",
           "Look for OpenCL functions in system header files") \
   do |v|
     options[:"system-headers"] = v
@@ -807,6 +956,6 @@ elsif options[:mode] == :"update-database" then
 elsif options[:mode] == :database then
   IcdGenerator.generate_from_database(options[:database])
 else
-  raise "Mode must be one of generate, database or update-database not #{options[:mode]}" 
+  raise "Mode must be one of generate, database or update-database not #{options[:mode]}"
 end
 
